import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision, save_list_of_lists, access_txt
from qs_mps.gs_multiprocessing import ground_state_ising

# DENSITY MATRIX RENORMALIZATION GROUP to find ground states of the 
# 1D Ising Transverse Field model changing the transverse field parameters

parser = argparse.ArgumentParser(prog="gs_search_Ising")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h (external transverse field)", type=float
)
parser.add_argument(
    "h_f", help="Final value of h (external transverse field)", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Spin chain lengths", nargs="*", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d","--dimension", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="trapezoidal", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="Ising", type=str
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=8,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-10,
    type=float,
)
parser.add_argument(
    "-fid",
    "--fidelity",
    help="Fidelity with exact solution. Doable only for small Ising chains",
    action="store_true",
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval = np.linspace(args.h_i, args.h_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/0_ISING"
    path_tensor = "D:/code/projects/0_ISING"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm98/Desktop/projects/0_ISING"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/0_ISING"
    path_tensor = "/Users/fradm/Desktop/projects/0_ISING"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

path = f"{parent_path}/results/tensors"
num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num)

Z = np.array([[1, 0], [0, -1]])

# ---------------------------------------------------------
# DMRG
# ---------------------------------------------------------
for L in args.Ls:
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False
    for chi in args.chis:  # L // 2 + 1
        args_mps = {
            "L": L,
            "d": args.dimension,
            "model": args.model,
            "chi": chi,
            "path": path_tensor,
            "type_shape": args.type_shape,
            "precision": precision,
            "trunc_chi": True,
            "trunc_tol": False,
            "where": args.where,
            "bond": args.bond,
            "J": 1,
            "eps": 0,
        }
        if __name__ == "__main__":
            
            
            energy_chi = []
            entropy_chi = []
            schmidt_vals_chi = []
            for J in interval:
                init_state = np.zeros((1, 2, 1))
                init_state[0, 0, 0] = 1
                init_tensor = [init_state for _ in range(L)]
                for h in interval:
                    precision = args_mps["precision"]
                    chain = MPS(
                        L=args_mps["L"],
                        d=args_mps["d"],
                        model=args_mps["model"],
                        chi=args_mps["chi"],
                        h=h,
                        J=J,
                        eps=args_mps["eps"],
                    )
                    chain.sites = init_tensor
                    chain.enlarge_chi(type_shape="rectangular", prnt=False)
                    # chain._random_state(seed=7, chi=args_mps["chi"], type_shape=args_mps["type_shape"])
                    chain.canonical_form(trunc_chi=False, trunc_tol=True)
                    # total
                    chain.order_param(op=Z)
                    mag = np.real(chain.mpo_first_moment())
                    if mag < 0:
                        chain.flipping_all()

                    energy, entropy, schmidt_vals = chain.DMRG(
                        trunc_tol=args_mps["trunc_tol"],
                        trunc_chi=args_mps["trunc_chi"],
                        where=args_mps["where"],
                        bond=args_mps["bond"],
                    )

                    print(f"energy of h:{h:.{precision}f}, J:{J:.{precision}f} is:\n {energy}")
                    print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")

                    chain.save_sites(path=args_mps["path"], precision=args_mps["precision"])
                    energy_chi.append(energy)
                    entropy_chi.append(entropy)
                    schmidt_vals_chi.append(schmidt_vals)

            # energy_chi, entropy_chi, schmidt_vals_chi = ground_state_ising(
            #     args_mps=args_mps, multpr=False, param=interval
            # )

            if args.bond == False:
                args.where = "all"

            energy_chi = np.array(energy_chi)
            energy_chi = np.array_split(energy_chi, args.npoints)
            save_list_of_lists(
                f"{parent_path}/results/energy_data/energies_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
            entropy_chi = np.array(entropy_chi)
            entropy_chi = np.array_split(entropy_chi, args.npoints)
            save_list_of_lists(
                f"{parent_path}/results/entropy/{args.where}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                entropy_chi,
            )
            schmidt_vals_chi = np.array(schmidt_vals_chi)
            schmidt_vals_chi = np.array_split(schmidt_vals_chi, args.npoints)
            save_list_of_lists(
                f"{parent_path}/results/entropy/{args.where}_schmidt_values_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                schmidt_vals_chi,
            )
            if args.where == "all":
                entropy_mid = access_txt(
                    f"{parent_path}/results/entropy/{args.where}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    L // 2,
                )
                np.savetxt(
                    f"{parent_path}/results/entropy/{L // 2}_bond_entropy_{args.model}_L_{L}_h_{args.h_i}-{args.h_f}_delta_{args.npoints}_chi_{chi}",
                    entropy_mid,
                )



import argparse
import numpy as np
from qs_mps.mps_class import MPS
from qs_mps.utils import get_precision, save_list_of_lists, access_txt, tensor_shapes
from qs_mps.applications.ANNNI.ground_state_multiprocessing import ground_state_ANNNI
from scipy.sparse.linalg._eigen.arpack.arpack import ArpackNoConvergence

# DENSITY MATRIX RENORMALIZATION GROUP to find ground states of the Z2 Pure Gauge Theory 
# changing the transverse field parameters in its dual formulation

parser = argparse.ArgumentParser(prog="gs_search_ANNNI")
parser.add_argument(
    "npoints",
    help="Number of points in an interval of transverse field values",
    type=int,
)
parser.add_argument(
    "h_i", help="Starting value of h external transverse field", type=float
)
parser.add_argument(
    "h_f", help="Final value of h external transverse field", type=float
)
parser.add_argument(
    "k_i", help="Starting value of J2 next nearest interaction coefficient", type=float
)
parser.add_argument(
    "k_f", help="Final value of J2 next nearest interaction coefficient", type=float
)
parser.add_argument(
    "path",
    help="Path to the drive depending on the device used. Available are 'pc', 'mac', 'marcos'",
    type=str,
)
parser.add_argument("-L", "--Ls", help="Spin chain lengths", nargs="*", type=int)
parser.add_argument("-D", "--chis", help="Simulated bond dimensions", nargs="+", type=int)
parser.add_argument("-d","--dimension", help="Physical dimension. By default 2", default=2, type=int)
parser.add_argument(
    "-ty", "--type_shape", help="Type of shape of the bond dimension. Available are: 'trapezoidal', 'pyramidal', 'rectangular'", default="rectangular", type=str
)
parser.add_argument(
    "-m", "--model", help="Model to simulate", default="ANNNI", type=str
)
parser.add_argument(
    "-mu", "--multpr", help="If True computes ground states with multiprocessing. By default False", action="store_true"
)
parser.add_argument(
    "-s",
    "--number_sweeps",
    help="Number of sweeps during the compression algorithm for each trotter step",
    default=8,
    type=int,
)
parser.add_argument(
    "-cv",
    "--conv_tol",
    help="Convergence tolerance of the compression algorithm",
    default=1e-10,
    type=float,
)
parser.add_argument(
    "-b",
    "--bond",
    help="Save the schmidt values for one bond. If False save for each bond. By default True",
    action="store_false",
)
parser.add_argument(
    "-w",
    "--where",
    help="Bond where we want to observe the Schmidt values, should be between 1 and (L-1)",
    default=-1,
    type=int,
)
parser.add_argument(
    "-v",
    "--save",
    help="Save the tensors. By default True",
    action="store_false",
)
parser.add_argument(
    "-tr",
    "--training",
    help="Save all the energies also the ones during the variational optimization. By default False",
    action="store_true",
)

args = parser.parse_args()

# define the interval of equally spaced values of external field
interval_h = np.linspace(args.h_i, args.h_f, args.npoints)
interval_k = np.linspace(args.k_i, args.k_f, args.npoints)

# take the path and precision to save files
# if we want to save the tensors we save them locally because they occupy a lot of memory
if args.path == "pc":
    parent_path = "G:/My Drive/projects/2_ANNNI"
    path_tensor = "D:/code/projects/2_ANNNI"
elif args.path == "mac":
    parent_path = "/Users/fradm98/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm98/Desktop/projects/2_ANNNI"
elif args.path == "marcos":
    parent_path = "/Users/fradm/Google Drive/My Drive/projects/2_ANNNI"
    path_tensor = "/Users/fradm/Desktop/projects/2_ANNNI"
else:
    raise SyntaxError("Path not valid. Choose among 'pc', 'mac', 'marcos'")

num = (args.h_f - args.h_i) / args.npoints
precision = get_precision(num)

Z = np.array([[1, 0], [0, -1]])
# ---------------------------------------------------------
# DMRG
# ---------------------------------------------------------
for L in args.Ls:
    if args.where == -1:
        args.where = L // 2
    elif args.where == -2:
        args.bond = False
    for chi in args.chis:  # L // 2 + 1
        args_mps = {
            "L": L,
            "d": args.dimension,
            "chi": chi,
            "type_shape": args.type_shape,
            "model": args.model,
            "trunc_tol": False,
            "trunc_chi": True,
            "where": args.where,
            "bond": args.bond,
            "path": path_tensor,
            "save": args.save,
            "precision": precision,
            "eps": 0,
        }
        if __name__ == "__main__":
            energy_chi = []
            entropy_chi = []
            schmidt_vals_chi = []
            for h in interval_h:
                # init_tensor = MPS(
                #         L=args_mps["L"],
                #         d=args_mps["d"],
                #         model="Ising",
                #         chi=args_mps["chi"],
                #         h=h,
                #         J=1,
                #         eps=args_mps["eps"],
                #         )
                # init_tensor.load_sites(path=path_tensor.rsplit("/", 1)[:-1][0]+"0_ISING", precision=args_mps["precision"])
                # init = init_tensor.sites.copy()
                for k in interval_k:
                    precision = args_mps["precision"]
                    chain = MPS(
                        L=args_mps["L"],
                        d=args_mps["d"],
                        model=args_mps["model"],
                        chi=args_mps["chi"],
                        h=h,
                        k=k,
                        J=1,
                        eps=args_mps["eps"],
                    )
                    chain._random_state(seed=3,type_shape="rectangular", chi=chi)
                    # chain._random_state(seed=7, chi=args_mps["chi"], type_shape=args_mps["type_shape"])
                    chain.canonical_form(trunc_chi=False, trunc_tol=True)
                    # # total
                    # chain.order_param(op=Z)
                    # mag = np.real(chain.mpo_first_moment())
                    # if mag < 0:
                    #     chain.flipping_all()

                    try:
                        energy, entropy, schmidt_vals = chain.DMRG(
                            trunc_tol=args_mps["trunc_tol"],
                            trunc_chi=args_mps["trunc_chi"],
                            where=args_mps["where"],
                            bond=args_mps["bond"],
                        )

                        print(f"energy of h:{h:.{precision}f}, k:{k:.{precision}f} is:\n {energy}")
                        print(f"Schmidt values in the middle of the chain:\n {schmidt_vals}")

                        chain.save_sites(path=args_mps["path"], precision=args_mps["precision"])
                        
                    except ArpackNoConvergence:
                        energy = np.nan
                        entropy = np.nan
                        schmidt_vals = np.nan
                        
                    energy_chi.append(energy)
                    entropy_chi.append(entropy)
                    schmidt_vals_chi.append(schmidt_vals)
        # energy_chi, entropy_chi, schmidt_vals_chi = ground_state_ANNNI(
        #     args_mps=args_mps, multpr=args.multpr, param=[interval_h,interval_k]
        # )

        if args.bond == False:
            args.where = "all"

        if args.training:
            save_list_of_lists(
                f"{parent_path}/results/energy_data/energies_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
            energy_gs = access_txt(
                    f"{parent_path}/results/energy_data/energies_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                    -1,
                )
            np.savetxt(
                f"{parent_path}/results/energy_data/energies_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                energy_gs,
            )
        else:
            np.savetxt(
                f"{parent_path}/results/energy_data/energies_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                energy_chi,
            )
             
        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
            entropy_chi,
        )

        save_list_of_lists(
            f"{parent_path}/results/entropy_data/{args.where}_schmidt_vals_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
            schmidt_vals_chi,
        )

        if args.where == "all":
            entropy_mid = access_txt(
                f"{parent_path}/results/entropy_data/{args.where}_bond_entropy_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                (L-1) // 2,
            )
            np.savetxt(
                f"{parent_path}/results/entropy_data/{L // 2}_bond_entropy_{args.model}_spin_{L}_h_{args.h_i}-{args.h_f}_k_{args.k_i}-{args.k_f}_delta_{args.npoints}_chi_{chi}",
                entropy_mid,
            )


import numpy as np
from scipy import sparse
from scipy.sparse.linalg import eigsh, expm_multiply
from .utils import get_middle_chain_schmidt_values, von_neumann_entropy, mps_to_vector


# -----------------------------------------------
# Sparse Pauli matrices
# -----------------------------------------------
def sparse_non_diag_paulis_indices(n, N):
    """
    Returns a tuple (row_indices, col_indices) containing the row and col indices of the non_zero elements
    of the tensor product of a non diagonal pauli matrix (x, y) acting over a single qubit in a Hilbert
    space of N qubits

    """
    if 0 <= n < N:
        block_length = 2 ** (N - n - 1)
        nblocks = 2**n
        ndiag_elements = block_length * nblocks
        k = np.arange(ndiag_elements, dtype=int)
        red_row_col_ind = (k % block_length) + 2 * (k // block_length) * block_length
        upper_diag_row_indices = red_row_col_ind
        upper_diag_col_indices = block_length + red_row_col_ind
        row_indices = np.concatenate((upper_diag_row_indices, upper_diag_col_indices))
        col_indices = np.concatenate((upper_diag_col_indices, upper_diag_row_indices))
        return row_indices, col_indices
    else:
        raise ValueError("Index n must fulfill 0 <= n < N")


def sparse_pauli_x(n, L, row_indices_cache=None, col_indices_cache=None):
    """
    Returns a CSC sparse matrix representation of the pauli_x matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        if (row_indices_cache is None) or (col_indices_cache is None):
            row_indices_cache, col_indices_cache = sparse_non_diag_paulis_indices(n, L)
        data = np.ones_like(row_indices_cache)
        result = sparse.csc_array(
            (data, (row_indices_cache, col_indices_cache)), shape=(2**L, 2**L)
        )  # , dtype=complex
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")


def sparse_pauli_y(n, L, row_indices_cache=None, col_indices_cache=None):
    """
    Returns a CSC sparse matrix representation of the pauli_y matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        if (row_indices_cache is None) or (col_indices_cache is None):
            row_indices_cache, col_indices_cache = sparse_non_diag_paulis_indices(n, L)
        data = -1j * np.ones_like(row_indices_cache)
        data[len(data) // 2 : :] = 1j
        result = sparse.csc_array(
            (data, (row_indices_cache, col_indices_cache)),
            shape=(2**L, 2**L),
            dtype=complex,
        )
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")


def sparse_pauli_z(n, L):
    """
    Returns a CSC sparse matrix representation of the pauli_z matrix acting over qubit n in a Hilbert space of L qubits
    0 <= n < L

    """
    if 0 <= n < L:
        block_length = 2 ** (L - n)
        nblocks = 2**n
        block = np.ones(block_length, dtype=int)
        block[block_length // 2 : :] = -1
        diag = np.tile(block, nblocks)
        row_col_indices = np.arange(2**L, dtype=int)
        result = sparse.csc_array(
            (diag, (row_col_indices, row_col_indices)),
            shape=(2**L, 2**L),
            dtype=complex,
        )
        return result
    else:
        raise ValueError("Index n must fulfill 0 <= n < L")


# -----------------------------------------------
# Sparse Ising Hamiltonian
# -----------------------------------------------
def sparse_ising_hamiltonian(J, h_t, h_l, L):
    """
    Returns a sparse representation of the Hamiltonian of the 1D Heisemberg model in a chain of length L
    with periodic boundary conditions (hbar = 1)
    J < 0: Antiferromagnetic case (Unique ground state of total angular momentum S=0)
    J > 0: Ferromagnetic case (L+1-fold degeneracy of the ground state of angular momentum L/2) -> Dicke states for even L
    """
    hamiltonian_l = sparse.csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_t = sparse.csc_array((2**L, 2**L), dtype=complex)
    hamiltonian_int = sparse.csc_array((2**L, 2**L), dtype=complex)

    # First sum over the terms containing sigma_x, sigma_y because the non-zero element indices are the same
    # so that this improves performance
    if h_t != 0:
        for n in range(L):
            n_row_indices, n_col_indices = sparse_non_diag_paulis_indices(n, L)
            n_pauli_x = sparse_pauli_x(n, L, n_row_indices, n_col_indices)
            hamiltonian_t += n_pauli_x
    if h_l != 0:
        for n in range(L):
            n_pauli_z = sparse_pauli_z(n, L)
            hamiltonian_l += n_pauli_z

    # Sum over sigma_z terms

    for n in range(L - 1):
        n_pauli_z = sparse_pauli_z(n, L)
        np1_pauli_z = sparse_pauli_z(n + 1, L)
        hamiltonian_int += n_pauli_z @ np1_pauli_z

    return -J * hamiltonian_int - h_t * hamiltonian_t - h_l * hamiltonian_l


# ---------------------------------------------------------------------------------------
# Sparse Ground state
# ---------------------------------------------------------------------------------------
def sparse_ising_ground_state(
    L: int, h_t: float, h_l: float = 1e-7, J: float = 1, k: int = 1
) -> sparse.csc_array:
    """
    exact_initial_state

    This function is computing the initial state given by an Ising Hamiltonian.

    L: int - chain size
    h_t: float - initial transverse field parameter
    h_l: float - initial longitudinal field parameter
    k: int - number of eigenvalues we want to compute. By default 1

    """
    print("Finding the Hamiltonian...")
    H = sparse_ising_hamiltonian(J=J, h_t=h_t, h_l=h_l, L=L)
    print("Hamiltonian found")
    e, v = eigsh(H, k=k, which="SA")
    print(f"first {k} eigenvalue(s) SA (Smallest (algebraic) eigenvalues): {e}")
    psi = v[:, 0]
    return psi


# ---------------------------------------------------------------------------------------
# Sparse U Evolution
# ---------------------------------------------------------------------------------------
def U_evolution_sparse(
    psi_init: sparse.csc_array,
    H_ev: sparse.csc_array,
    trotter: int,
    time: float,
):
    """
    U_evolution

    This function applies a time evolution operator to some initial state.
    The evolution operator uses the Ising hamiltonian with some tunable parameters.

    psi_init: csc_array - initial state to be evolved
    H_ev: csc_array - Time evolution Hamiltonian
    trotter: int - the trotter step we are during the evolution
    time: float - indicates the final time we want to reach

    """
    delta = time / trotter
    H_ev = -1j * delta * H_ev
    psi_ev = expm_multiply(H_ev, psi_init)
    return psi_ev


# ---------------------------------------------------------------------------------------
# Sparse exact Evolution
# ---------------------------------------------------------------------------------------
def exact_evolution_sparse(
    L: int,
    h_t: float,
    h_ev: float,
    time: float,
    trotter_steps: int,
    h_l: float = 1e-7,
    flip: bool = False,
    where: int = -1,
    bond: bool = True,
    model: str = None,
):
    """
    exact_evolution

    This function evolves an initial state for trotter steps times.
    We extract at each time step the local and total magnetization.

    L: int - chain size
    trotter_step: int - indicates the specific trotter step we are evolving
    time: float - indicates the final time we want to reach
    h_t: float - initial transverse field parameter
    h_l: float - initial longitudinal field parameter
    h_ev: float - evolution transverse field parameter
    where: int - in which bond we perform the Schmidt decompositon
    bond: bool - if we compute the entropy for all the bonds. By default False

    """
    if bond:
        entropy = [0]
    else:
        entropy = [0] * (L - 3)

    # psi_exact = sparse_ising_ground_state(L=L, h_t=h_t, h_l=h_l, k=6).reshape(2**L, 1)
    # for now only all spin up
    # psi_exact = np.zeros((2**L, 1))
    # psi_exact[0] = 1
    init_state = np.zeros((1, 2, 1))
    init_state[0, 0, 0] = 1
    mps = [init_state for _ in range(L)]
    psi_exact = mps_to_vector(mps).reshape(2**L, 1)
    if flip:
        flip = sparse_pauli_x(n=L // 2, L=L)
        psi_exact = sparse.csc_array(flip @ psi_exact)
    # local Z
    mag_loc_Z_op = [sparse_pauli_z(n=i, L=L) for i in range(L)]
    # local X
    mag_loc_X_op = sparse_pauli_x(n=L // 2, L=L)
    # total
    mag_tot_op = sparse_ising_hamiltonian(L=L, h_t=0, h_l=-1, J=0)

    mag_exact_loc_Z = []
    mag_exact_loc_X = []
    mag_exact_tot = []

    # local Z
    mag_exact = []
    for i in range(L):
        mag_exact.append(
            (psi_exact.T.conjugate() @ mag_loc_Z_op[i] @ psi_exact)[0, 0].real
        )
    mag_exact_loc_Z.append(mag_exact)

    # local X
    mag_exact_loc_X.append(
        (psi_exact.T.conjugate() @ mag_loc_X_op @ psi_exact)[0, 0].real
    )

    # total
    mag_exact_tot.append((psi_exact.T.conjugate() @ mag_tot_op @ psi_exact)[0, 0].real)

    H_ev = sparse_ising_hamiltonian(L=L, J=1, h_l=h_l, h_t=h_ev)

    entropy_tot = []
    entropy_tot.append(entropy)

    psi_new = psi_exact
    for trott in range(trotter_steps):
        print(f"-------- Trotter step {trott} ---------")
        # exact
        psi_new = U_evolution_sparse(
            psi_init=psi_new, H_ev=H_ev, trotter=trotter_steps, time=time
        )
        # entropy
        sing_vals = get_middle_chain_schmidt_values(psi_new, where=where, bond=bond)
        entropy = []
        for s in sing_vals:
            ent = von_neumann_entropy(s)
            entropy.append(ent)
        entropy_tot.append(entropy)
        # local Z
        mag_exact = []
        for i in range(L):
            mag_exact.append(
                (psi_new.T.conjugate() @ mag_loc_Z_op[i] @ psi_new)[0, 0].real
            )
        mag_exact_loc_Z.append(mag_exact)

        # local X
        mag_exact_loc_X.append(
            (psi_new.T.conjugate() @ mag_loc_X_op @ psi_new)[0, 0].real
        )

        # total
        mag_exact_tot.append((psi_new.T.conjugate() @ mag_tot_op @ psi_new)[0, 0].real)
    return psi_new, mag_exact_loc_Z, mag_exact_loc_X, mag_exact_tot, entropy_tot


# psi_new, mag_exact_loc, mag_exact_loc_X, mag_exact_tot, entropy_tot = exact_evolution_sparse(L=15, h_t=0, h_ev=0.3, time=10, trotter_steps=5, flip=True, where=7)
